@@ -13,7 +13,6 @@ import java.util.List;
 import java.util.Map;
 import java.util.NoSuchElementException;
 import java.util.Set;
-
 import javax.annotation.Nonnull;
 
 import org.bouncycastle.openpgp.PGPPublicKey;
@@ -21,8 +20,11 @@ import org.bouncycastle.openpgp.PGPPublicKeyRing;
 import org.bouncycastle.openpgp.PGPPublicKeyRingCollection;
 import org.bouncycastle.openpgp.operator.PBEKeyEncryptionMethodGenerator;
 import org.bouncycastle.openpgp.operator.PGPKeyEncryptionMethodGenerator;
+import org.bouncycastle.openpgp.operator.PublicKeyKeyEncryptionMethodGenerator;
 import org.pgpainless.algorithm.EncryptionPurpose;
 import org.pgpainless.algorithm.SymmetricKeyAlgorithm;
+import org.pgpainless.authentication.CertificateAuthenticity;
+import org.pgpainless.authentication.CertificateAuthority;
 import org.pgpainless.exception.KeyException;
 import org.pgpainless.implementation.ImplementationFactory;
 import org.pgpainless.key.OpenPgpFingerprint;
@@ -34,7 +36,7 @@ import org.pgpainless.util.Passphrase;
 /**
  * Options for the encryption process.
  * This class can be used to set encryption parameters, like encryption keys and passphrases, algorithms etc.
- *
+ * <p>
  * A typical use might look like follows:
  * <pre>
  * {@code
@@ -43,11 +45,11 @@ import org.pgpainless.util.Passphrase;
  * opt.addPassphrase(Passphrase.fromPassword("AdditionalDecryptionPassphrase123"));
  * }
  * </pre>
- *
+ *<p>
  * To use a custom symmetric encryption algorithm, use {@link #overrideEncryptionAlgorithm(SymmetricKeyAlgorithm)}.
  * This will cause PGPainless to use the provided algorithm for message encryption, instead of negotiating an algorithm
  * by inspecting the provided recipient keys.
- *
+ * <p>
  * By default, PGPainless will encrypt to all suitable, encryption capable subkeys on each recipient's certificate.
  * This behavior can be changed per recipient, e.g. by calling
  * <pre>
@@ -65,6 +67,8 @@ public class EncryptionOptions {
     private final Map<SubkeyIdentifier, KeyRingInfo> keyRingInfo = new HashMap<>();
     private final Map<SubkeyIdentifier, KeyAccessor> keyViews = new HashMap<>();
     private final EncryptionKeySelector encryptionKeySelector = encryptToAllCapableSubkeys();
+    private boolean allowEncryptionWithMissingKeyFlags = false;
+    private Date evaluationDate = new Date();
 
     private SymmetricKeyAlgorithm encryptionAlgorithmOverride = null;
 
@@ -76,7 +80,7 @@ public class EncryptionOptions {
         this(EncryptionPurpose.ANY);
     }
 
-    public EncryptionOptions(EncryptionPurpose purpose) {
+    public EncryptionOptions(@Nonnull EncryptionPurpose purpose) {
         this.purpose = purpose;
     }
 
@@ -84,7 +88,7 @@ public class EncryptionOptions {
      * Factory method to create an {@link EncryptionOptions} object which will encrypt for keys
      * which carry either the {@link org.pgpainless.algorithm.KeyFlag#ENCRYPT_COMMS} or
      * {@link org.pgpainless.algorithm.KeyFlag#ENCRYPT_STORAGE} flag.
-     *
+     * <p>
      * Use this if you are not sure.
      *
      * @return encryption options
@@ -93,6 +97,17 @@ public class EncryptionOptions {
         return new EncryptionOptions();
     }
 
+    /**
+     * Override the evaluation date for recipient keys with the given date.
+     *
+     * @param evaluationDate new evaluation date
+     * @return this
+     */
+    public EncryptionOptions setEvaluationDate(@Nonnull Date evaluationDate) {
+        this.evaluationDate = evaluationDate;
+        return this;
+    }
+
     /**
      * Factory method to create an {@link EncryptionOptions} object which will encrypt for keys
      * which carry the flag {@link org.pgpainless.algorithm.KeyFlag#ENCRYPT_COMMS}.
@@ -113,13 +128,52 @@ public class EncryptionOptions {
         return new EncryptionOptions(EncryptionPurpose.STORAGE);
     }
 
+    /**
+     * Identify authenticatable certificates for the given user-ID by querying the {@link CertificateAuthority} for
+     * identifiable bindings.
+     * Add all acceptable bindings, whose trust amount is larger or equal to the target amount to the list of recipients.
+     * @param userId userId
+     * @param email if true, treat the user-ID as an email address and match all user-IDs containing the mail address
+     * @param authority certificate authority
+     * @return encryption options
+     */
+    public EncryptionOptions addAuthenticatableRecipients(String userId, boolean email, CertificateAuthority authority) {
+        return addAuthenticatableRecipients(userId, email, authority, 120);
+    }
+
+    /**
+     * Identify authenticatable certificates for the given user-ID by querying the {@link CertificateAuthority} for
+     * identifiable bindings.
+     * Add all acceptable bindings, whose trust amount is larger or equal to the target amount to the list of recipients.
+     * @param userId userId
+     * @param email if true, treat the user-ID as an email address and match all user-IDs containing the mail address
+     * @param authority certificate authority
+     * @param targetAmount target amount (120 = fully authenticated, 240 = doubly authenticated,
+     *                    60 = partially authenticated...)
+     * @return encryption options
+     */
+    public EncryptionOptions addAuthenticatableRecipients(String userId, boolean email, CertificateAuthority authority, int targetAmount) {
+        List<CertificateAuthenticity> identifiedCertificates = authority.lookupByUserId(userId, email, evaluationDate, targetAmount);
+        boolean foundAcceptable = false;
+        for (CertificateAuthenticity candidate : identifiedCertificates) {
+            if (candidate.isAuthenticated()) {
+                addRecipient(candidate.getCertificate());
+                foundAcceptable = true;
+            }
+        }
+        if (!foundAcceptable) {
+            throw new IllegalArgumentException("Could not identify any trust-worthy certificates for '" + userId + "' and target trust amount " + targetAmount);
+        }
+        return this;
+    }
+
     /**
      * Add all key rings in the provided {@link Iterable} (e.g. {@link PGPPublicKeyRingCollection}) as recipients.
      *
      * @param keys keys
      * @return this
      */
-    public EncryptionOptions addRecipients(Iterable<PGPPublicKeyRing> keys) {
+    public EncryptionOptions addRecipients(@Nonnull Iterable<PGPPublicKeyRing> keys) {
         if (!keys.iterator().hasNext()) {
             throw new IllegalArgumentException("Set of recipient keys cannot be empty.");
         }
@@ -155,7 +209,7 @@ public class EncryptionOptions {
      * @param userId user id
      * @return this
      */
-    public EncryptionOptions addRecipient(PGPPublicKeyRing key, String userId) {
+    public EncryptionOptions addRecipient(@Nonnull PGPPublicKeyRing key, @Nonnull CharSequence userId) {
         return addRecipient(key, userId, encryptionKeySelector);
     }
 
@@ -168,11 +222,13 @@ public class EncryptionOptions {
      * @param encryptionKeySelectionStrategy strategy to select one or more encryption subkeys to encrypt to
      * @return this
      */
-    public EncryptionOptions addRecipient(PGPPublicKeyRing key, String userId, EncryptionKeySelector encryptionKeySelectionStrategy) {
-        KeyRingInfo info = new KeyRingInfo(key, new Date());
+    public EncryptionOptions addRecipient(@Nonnull PGPPublicKeyRing key,
+                                          @Nonnull CharSequence userId,
+                                          @Nonnull EncryptionKeySelector encryptionKeySelectionStrategy) {
+        KeyRingInfo info = new KeyRingInfo(key, evaluationDate);
 
         List<PGPPublicKey> encryptionSubkeys = encryptionKeySelectionStrategy
-                .selectEncryptionSubkeys(info.getEncryptionSubkeys(userId, purpose));
+                .selectEncryptionSubkeys(info.getEncryptionSubkeys(userId.toString(), purpose));
         if (encryptionSubkeys.isEmpty()) {
             throw new KeyException.UnacceptableEncryptionKeyException(OpenPgpFingerprint.of(key));
         }
@@ -180,8 +236,8 @@ public class EncryptionOptions {
         for (PGPPublicKey encryptionSubkey : encryptionSubkeys) {
             SubkeyIdentifier keyId = new SubkeyIdentifier(key, encryptionSubkey.getKeyID());
             keyRingInfo.put(keyId, info);
-            keyViews.put(keyId, new KeyAccessor.ViaUserId(info, keyId, userId));
-            addRecipientKey(key, encryptionSubkey);
+            keyViews.put(keyId, new KeyAccessor.ViaUserId(info, keyId, userId.toString()));
+            addRecipientKey(key, encryptionSubkey, false);
         }
 
         return this;
@@ -193,7 +249,7 @@ public class EncryptionOptions {
      * @param key key ring
      * @return this
      */
-    public EncryptionOptions addRecipient(PGPPublicKeyRing key) {
+    public EncryptionOptions addRecipient(@Nonnull PGPPublicKeyRing key) {
         return addRecipient(key, encryptionKeySelector);
     }
 
@@ -204,10 +260,36 @@ public class EncryptionOptions {
      * @param encryptionKeySelectionStrategy strategy used to select one or multiple encryption subkeys.
      * @return this
      */
-    public EncryptionOptions addRecipient(PGPPublicKeyRing key, EncryptionKeySelector encryptionKeySelectionStrategy) {
-        Date evaluationDate = new Date();
-        KeyRingInfo info;
-        info = new KeyRingInfo(key, evaluationDate);
+    public EncryptionOptions addRecipient(@Nonnull PGPPublicKeyRing key,
+                                          @Nonnull EncryptionKeySelector encryptionKeySelectionStrategy) {
+        return addAsRecipient(key, encryptionKeySelectionStrategy, false);
+    }
+
+    /**
+     * Add a certificate as hidden recipient.
+     * The recipients key-id will be obfuscated by setting a wildcard key ID.
+     *
+     * @param key recipient key
+     * @return this
+     */
+    public EncryptionOptions addHiddenRecipient(@Nonnull PGPPublicKeyRing key) {
+        return addHiddenRecipient(key, encryptionKeySelector);
+    }
+
+    /**
+     * Add a certificate as hidden recipient, using the provided {@link EncryptionKeySelector} to select recipient subkeys.
+     * The recipients key-ids will be obfuscated by setting a wildcard key ID instead.
+     *
+     * @param key recipient key
+     * @param encryptionKeySelectionStrategy strategy to select recipient (sub) keys.
+     * @return this
+     */
+    public EncryptionOptions addHiddenRecipient(PGPPublicKeyRing key, EncryptionKeySelector encryptionKeySelectionStrategy) {
+        return addAsRecipient(key, encryptionKeySelectionStrategy, true);
+    }
+
+    private EncryptionOptions addAsRecipient(PGPPublicKeyRing key, EncryptionKeySelector encryptionKeySelectionStrategy, boolean wildcardKeyId) {
+        KeyRingInfo info = new KeyRingInfo(key, evaluationDate);
 
         Date primaryKeyExpiration;
         try {
@@ -221,6 +303,23 @@ public class EncryptionOptions {
 
         List<PGPPublicKey> encryptionSubkeys = encryptionKeySelectionStrategy
                 .selectEncryptionSubkeys(info.getEncryptionSubkeys(purpose));
+
+        // There are some legacy keys around without key flags.
+        // If we allow encryption for those keys, we add valid keys without any key flags, if they are
+        // capable of encryption by means of their algorithm
+        if (encryptionSubkeys.isEmpty() && allowEncryptionWithMissingKeyFlags) {
+            List<PGPPublicKey> validSubkeys = info.getValidSubkeys();
+            for (PGPPublicKey validSubkey : validSubkeys) {
+                if (!validSubkey.isEncryptionKey()) {
+                    continue;
+                }
+                // only add encryption keys with no key flags.
+                if (info.getKeyFlagsOf(validSubkey.getKeyID()).isEmpty()) {
+                    encryptionSubkeys.add(validSubkey);
+                }
+            }
+        }
+
         if (encryptionSubkeys.isEmpty()) {
             throw new KeyException.UnacceptableEncryptionKeyException(OpenPgpFingerprint.of(key));
         }
@@ -229,16 +328,19 @@ public class EncryptionOptions {
             SubkeyIdentifier keyId = new SubkeyIdentifier(key, encryptionSubkey.getKeyID());
             keyRingInfo.put(keyId, info);
             keyViews.put(keyId, new KeyAccessor.ViaKeyId(info, keyId));
-            addRecipientKey(key, encryptionSubkey);
+            addRecipientKey(key, encryptionSubkey, wildcardKeyId);
         }
 
         return this;
     }
 
-    private void addRecipientKey(PGPPublicKeyRing keyRing, PGPPublicKey key) {
+    private void addRecipientKey(@Nonnull PGPPublicKeyRing keyRing,
+                                 @Nonnull PGPPublicKey key,
+                                 boolean wildcardKeyId) {
         encryptionKeys.add(new SubkeyIdentifier(keyRing, key.getKeyID()));
-        PGPKeyEncryptionMethodGenerator encryptionMethod = ImplementationFactory
+        PublicKeyKeyEncryptionMethodGenerator encryptionMethod = ImplementationFactory
                 .getInstance().getPublicKeyKeyEncryptionMethodGenerator(key);
+        encryptionMethod.setUseWildcardKeyID(wildcardKeyId);
         addEncryptionMethod(encryptionMethod);
     }
 
@@ -248,7 +350,7 @@ public class EncryptionOptions {
      * @param passphrase passphrase
      * @return this
      */
-    public EncryptionOptions addPassphrase(Passphrase passphrase) {
+    public EncryptionOptions addPassphrase(@Nonnull Passphrase passphrase) {
         if (passphrase.isEmpty()) {
             throw new IllegalArgumentException("Passphrase must not be empty.");
         }
@@ -268,7 +370,7 @@ public class EncryptionOptions {
      * @param encryptionMethod encryption method
      * @return this
      */
-    public EncryptionOptions addEncryptionMethod(PGPKeyEncryptionMethodGenerator encryptionMethod) {
+    public EncryptionOptions addEncryptionMethod(@Nonnull PGPKeyEncryptionMethodGenerator encryptionMethod) {
         encryptionMethods.add(encryptionMethod);
         return this;
     }
@@ -304,7 +406,7 @@ public class EncryptionOptions {
      * @param encryptionAlgorithm encryption algorithm override
      * @return this
      */
-    public EncryptionOptions overrideEncryptionAlgorithm(SymmetricKeyAlgorithm encryptionAlgorithm) {
+    public EncryptionOptions overrideEncryptionAlgorithm(@Nonnull SymmetricKeyAlgorithm encryptionAlgorithm) {
         if (encryptionAlgorithm == SymmetricKeyAlgorithm.NULL) {
             throw new IllegalArgumentException("Plaintext encryption can only be used to denote unencrypted secret keys.");
         }
@@ -312,8 +414,31 @@ public class EncryptionOptions {
         return this;
     }
 
+    /**
+     * If this method is called, subsequent calls to {@link #addRecipient(PGPPublicKeyRing)} will allow encryption
+     * for subkeys that do not carry any {@link org.pgpainless.algorithm.KeyFlag} subpacket.
+     * This is a workaround for dealing with legacy keys that have no key flags subpacket but rely on the key algorithm
+     * type to convey the subkeys use.
+     *
+     * @return this
+     */
+    public EncryptionOptions setAllowEncryptionWithMissingKeyFlags() {
+        this.allowEncryptionWithMissingKeyFlags = true;
+        return this;
+    }
+
+    /**
+     * Return <pre>true</pre> iff the user specified at least one encryption method,
+     * <pre>false</pre> otherwise.
+     *
+     * @return encryption methods is not empty
+     */
+    public boolean hasEncryptionMethod() {
+        return !encryptionMethods.isEmpty();
+    }
+
     public interface EncryptionKeySelector {
-        List<PGPPublicKey> selectEncryptionSubkeys(List<PGPPublicKey> encryptionCapableKeys);
+        List<PGPPublicKey> selectEncryptionSubkeys(@Nonnull List<PGPPublicKey> encryptionCapableKeys);
     }
 
     /**
@@ -324,7 +449,7 @@ public class EncryptionOptions {
     public static EncryptionKeySelector encryptToFirstSubkey() {
         return new EncryptionKeySelector() {
             @Override
-            public List<PGPPublicKey> selectEncryptionSubkeys(List<PGPPublicKey> encryptionCapableKeys) {
+            public List<PGPPublicKey> selectEncryptionSubkeys(@Nonnull List<PGPPublicKey> encryptionCapableKeys) {
                 return encryptionCapableKeys.isEmpty() ? Collections.emptyList() : Collections.singletonList(encryptionCapableKeys.get(0));
             }
         };
@@ -338,7 +463,7 @@ public class EncryptionOptions {
     public static EncryptionKeySelector encryptToAllCapableSubkeys() {
         return new EncryptionKeySelector() {
             @Override
-            public List<PGPPublicKey> selectEncryptionSubkeys(List<PGPPublicKey> encryptionCapableKeys) {
+            public List<PGPPublicKey> selectEncryptionSubkeys(@Nonnull List<PGPPublicKey> encryptionCapableKeys) {
                 return encryptionCapableKeys;
             }
         };
